"""Reduction algorithms."""

from typing import Tuple, Mapping, Callable, Optional, Any, cast
from typing_extensions import TypedDict

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore

from . import accel
from . import tune
from .abc import AbstractContext, AbstractCommandQueue


_TuningDict = TypedDict('_TuningDict', {'wgsx': int, 'wgsy': int})


class HReduceTemplate:
    """Performs reduction along rows in a 2D array.

    Only commutative reduction operators are supported.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    dtype
        Type of data elements
    ctype
        Type (in C/CUDA, not numpy) of data elements
    op
        C expression to combine two variables, *a* and *b*
    identity
        C expression for an identity value for *op*
    extra_code
        Arbitrary C code to paste in (for use by *op* or *identity*)
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of workitems per workgroup per row
        - wgsy: number of rows to handle per workgroup
    """

    def __init__(self, context: AbstractContext, dtype: DTypeLike, ctype: str,
                 op: str, identity: str, extra_code: str = '',
                 tuning: Optional[_TuningDict] = None) -> None:
        self.context = context
        self.dtype = dtype
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype, op, identity, extra_code)
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.op = op
        self.identity = identity
        self.extra_code = extra_code
        self.program = accel.build(context, 'hreduce.mako', {
            'wgsx': self.wgsx, 'wgsy': self.wgsy, 'type': self.ctype,
            'op': op, 'identity': identity, 'extra_code': extra_code})

    @classmethod
    @tune.autotuner(test={'wgsx': 64, 'wgsy': 4})
    def autotune(cls, context: AbstractContext, dtype: DTypeLike, ctype: str,
                 op: str, identity: str, extra_code: str) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        shape = (2048, 1024)
        src = accel.DeviceArray(context, shape, dtype=dtype)
        dest = accel.DeviceArray(context, (shape[0],), dtype=dtype)

        def generate(wgsx: int, wgsy: int) -> Callable[[int], float]:
            wgs = wgsx * wgsy
            if wgs < 32 or wgs > 1024:
                raise RuntimeError(f'Skipping work group size {wgsx}x{wgsy}')
            template = cls(context, dtype, ctype, op, identity, extra_code,
                           {'wgsx': wgsx, 'wgsy': wgsy})
            fn = template.instantiate(queue, shape)
            fn.bind(src=src, dest=dest)
            fn.ensure_all_bound()
            return tune.make_measure(queue, fn)

        return cast(_TuningDict, tune.autotune(generate,
                                               wgsx=[32, 64, 128],
                                               wgsy=[1, 2, 4, 8, 16]))

    def instantiate(self, command_queue: AbstractCommandQueue, shape: Tuple[int, int],
                    column_range: Optional[Tuple[int, int]] = None,
                    allocator: Optional[accel.AbstractAllocator] = None) -> 'HReduce':
        return HReduce(self, command_queue, shape, column_range, allocator)


class HReduce(accel.Operation):
    """Concrete instance of :class:`HReduceTemplate`.

    In each row, the elements in the specified column range are reduced using
    the reduction operator supplied to the template.

    .. rubric:: Slots

    **src** : *rows* × *columns*
        Input data
    **dest** : *rows*
        Output reductions

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    shape
        Shape for the source slot
    column_range
        Half-open range of columns to reduce (defaults to the entire array)
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: HReduceTemplate, command_queue: AbstractCommandQueue,
                 shape: Tuple[int, int], column_range: Optional[Tuple[int, int]] = None,
                 allocator: Optional[accel.AbstractAllocator] = None) -> None:
        if len(shape) != 2:
            raise ValueError('shape must be 2-dimensional')
        if column_range is None:
            column_range = (0, shape[1])
        if column_range[0] < 0 or column_range[1] > shape[1]:
            raise ValueError('column range overflows the array')
        if column_range[0] >= column_range[1]:
            raise ValueError('column range is empty')

        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('hreduce')
        self.column_range = column_range
        self.slots['src'] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy), shape[1]),
            self.template.dtype)
        self.slots['dest'] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy),),
            self.template.dtype)

    def _run(self) -> None:
        src = self.buffer('src')
        dest = self.buffer('dest')
        rows_padded = accel.roundup(src.shape[0], self.template.wgsy)
        n_columns = self.column_range[1] - self.column_range[0]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer, dest.buffer,
                np.int32(self.column_range[0]), np.int32(n_columns),
                np.int32(src.padded_shape[1])
            ],
            global_size=(self.template.wgsx, rows_padded),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['src'].shape,    # type: ignore
            'column_range': self.column_range,
            'op': self.template.op,
            'identity': self.template.identity,
            'extra_code': self.template.extra_code
        }



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for wg_reduce.mako and :mod:`katsdpsigproc.reduce`."""

import os
from typing import Tuple

import numpy as np
import pytest

from katsdpsigproc.accel import DeviceArray, build
from katsdpsigproc import reduce
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


@pytest.mark.parametrize('size', [1, 4, 12, 16, 32, 87, 97, 160, 256])
@pytest.mark.parametrize('allow_shuffle', [True, False])
@pytest.mark.parametrize('broadcast', [True, False])
@pytest.mark.parametrize(
    'kernel_name, op', [('test_reduce_add', np.add), ('test_reduce_max', np.maximum)])
def test_reduce(context: AbstractContext, command_queue: AbstractCommandQueue,
                size: int, allow_shuffle: bool, broadcast: bool,
                kernel_name: str, op: np.ufunc) -> None:
    rs = np.random.RandomState(seed=1)  # Fixed seed to make test repeatable
    rows = 256 // size
    data = rs.randint(0, 1000, size=(rows, size)).astype(np.int32)
    program = build(
        context, 'test_reduce.mako',
        {
            'size': size,
            'allow_shuffle': allow_shuffle,
            'broadcast': broadcast,
            'rows': rows
        },
        extra_dirs=[os.path.dirname(__file__)]
    )

    kernel = program.get_kernel(kernel_name)
    data_d = DeviceArray(context=context, shape=data.shape, dtype=data.dtype)
    data_d.set(command_queue, data)
    out_d = DeviceArray(context=context, shape=data.shape, dtype=data.dtype)
    dims = tuple(reversed(data.shape))
    command_queue.enqueue_kernel(
        kernel, [data_d.buffer, out_d.buffer], local_size=dims, global_size=dims)
    out = out_d.get(command_queue)
    expected = op.reduce(data, axis=1, keepdims=True)
    if broadcast:
        expected = np.repeat(expected, data.shape[1], axis=1)
    else:
        out = out[:, 0:1]
    np.testing.assert_array_equal(expected, out)


class TestHReduce:
    """Tests for :class:`katsdpsigproc.reduce.HReduce`."""

    @pytest.mark.parametrize(
        'rows, columns, column_range',
        [
            (129, 173, (67, 128)),
            (7, 8, (1, 7))  # Test that the special case for columns < work group size works
        ]
    )
    def test(self, context: AbstractContext, command_queue: AbstractCommandQueue,
             rows: int, columns: int, column_range: Tuple[int, int]) -> None:
        template = reduce.HReduceTemplate(context, np.uint32, 'unsigned int', 'a + b', '0')
        fn = template.instantiate(command_queue, (rows, columns), column_range)
        fn.ensure_all_bound()
        device_src = fn.buffer('src')
        device_dest = fn.buffer('dest')
        src = device_src.empty_like()
        rs = np.random.RandomState(seed=1)  # To be reproducible
        src[:] = rs.randint(0, 100000, (rows, columns))
        device_src.set(command_queue, src)
        fn()
        dest = device_dest.get(command_queue)
        expected = np.sum(src[:, column_range[0]:column_range[1]], axis=1)
        np.testing.assert_equal(expected, dest)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext) -> None:
        """Test that autotuner runs successfully."""
        reduce.HReduceTemplate(context, np.uint32, 'unsigned int', 'a + b', '0')



################################################################################
# Copyright (c) 2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

pytest_plugins = ["katsdpsigproc.pytest_plugin"]



#!/usr/bin/env python3

import os

import numpy as np

from katsdpsigproc.accel import Operation, IOSlot, create_some_context, build


class SumTemplate:
    def __init__(self, context):
        self.wgs = 128
        self.program = build(context, 'sum.mako', {'wgs': self.wgs},
                             extra_dirs=[os.path.dirname(__file__)])

    def instantiate(self, command_queue, size):
        return Sum(self, command_queue, size)


class Sum(Operation):
    def __init__(self, template, command_queue, size):
        if size % template.wgs:
            raise ValueError(f'size must be a multiple of {template.wgs}')
        super().__init__(command_queue)
        self.template = template
        self.slots['src'] = IOSlot((size,), np.int32)
        self.slots['dest'] = IOSlot((size // template.wgs,), np.int32)
        self.kernel = template.program.get_kernel('reduce')

    def _run(self):
        src = self.buffer('src')
        dest = self.buffer('dest')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [src.buffer, dest.buffer],
            global_size=src.shape,
            local_size=(self.template.wgs,)
        )


def main():
    ctx = create_some_context()
    queue = ctx.create_command_queue()
    op = SumTemplate(ctx).instantiate(queue, 1024)
    op.ensure_all_bound()
    src = np.random.randint(1, 100, size=op.buffer('src').shape).astype(np.int32)
    op.buffer('src').set(queue, src)
    op()
    dest = op.buffer('dest').get(queue)
    wgs = op.template.wgs
    expected = src.reshape(-1, wgs).sum(axis=1)
    np.testing.assert_equal(dest, expected)
    print(dest)


if __name__ == '__main__':
    main()



################################################################################
# Copyright (c) 2016-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.resource`."""

import time
import asyncio
import queue
import logging
from typing import List

import pytest

from katsdpsigproc import resource
from katsdpsigproc.abc import AbstractEvent


class TestWaitUntil:
    async def test_result(self) -> None:
        """wait_until returns before the timeout if a result is set."""
        future = asyncio.Future()     # type: asyncio.Future[int]
        loop = asyncio.get_event_loop()
        loop.call_later(0.1, future.set_result, 42)
        result = await resource.wait_until(future, loop.time() + 1000000)
        assert result == 42

    async def test_already_set(self) -> None:
        """wait_until returns if a future has a result set before the call."""
        future = asyncio.Future()     # type: asyncio.Future[int]
        future.set_result(42)
        loop = asyncio.get_event_loop()
        result = await resource.wait_until(future, loop.time() + 1000000)
        assert result == 42

    async def test_exception(self) -> None:
        """wait_until rethrows an exception set on the future."""
        future = asyncio.Future()     # type: asyncio.Future[int]
        loop = asyncio.get_event_loop()
        loop.call_later(0.1, future.set_exception, ValueError('test'))
        with pytest.raises(ValueError):
            await resource.wait_until(future, loop.time() + 1000000)

    async def test_timeout(self) -> None:
        """wait_until throws `asyncio.TimeoutError` if it times out, and cancels the future."""
        future = asyncio.Future()     # type: asyncio.Future[int]
        loop = asyncio.get_event_loop()
        with pytest.raises(asyncio.TimeoutError):
            await resource.wait_until(future, loop.time() + 0.01)
        assert future.cancelled()

    async def test_shield(self) -> None:
        """wait_until does not cancel the future if it is wrapped in shield."""
        future = asyncio.Future()     # type: asyncio.Future[int]
        loop = asyncio.get_event_loop()
        with pytest.raises(asyncio.TimeoutError):
            await resource.wait_until(asyncio.shield(future), loop.time() + 0.01)
        assert not future.cancelled()


class DummyEvent(AbstractEvent):
    """Dummy version of katsdpsigproc.accel event.

    The :meth:`wait` method just sleeps for a small time and then appends the
    event to a queue.
    """

    def __init__(self, completed: queue.Queue) -> None:
        self.complete = False
        self.completed = completed

    def wait(self) -> None:
        if not self.complete:
            time.sleep(0.1)
            self.completed.put(self)
            self.complete = True

    def time_since(self, prior_event: 'DummyEvent') -> float:
        return 0.0

    def time_till(self, next_event: 'DummyEvent') -> float:
        return 0.0


class TestResource:
    def setup_method(self) -> None:
        self.completed = queue.Queue()      # type: queue.Queue[resource.ResourceAllocation[int]]

    async def _run_frame(self, acq: resource.ResourceAllocation[int],
                         event: AbstractEvent) -> None:
        with acq as value:
            assert value == 42
            await acq.wait_events()
            acq.ready([event])
            self.completed.put(acq)

    async def test_wait_events(self) -> None:
        """Test :meth:`.ResourceAllocation.wait_events`."""
        r = resource.Resource(42)
        a0 = r.acquire()
        a1 = r.acquire()
        e0 = DummyEvent(self.completed)
        e1 = DummyEvent(self.completed)
        run1 = asyncio.ensure_future(self._run_frame(a1, e1))
        run0 = asyncio.ensure_future(self._run_frame(a0, e0))
        await run0
        await run1
        order = []
        try:
            while True:
                order.append(self.completed.get_nowait())
        except queue.Empty:
            pass
        assert order == [a0, e0, a1]

    async def test_context_manager_exception(self) -> None:
        """Test using :class:`resource.ResourceAllocation` as a \
        context manager when an error is raised."""
        r = resource.Resource(None)
        a0 = r.acquire()
        a1 = r.acquire()
        with pytest.raises(RuntimeError):
            with a0:
                await a0.wait_events()
                raise RuntimeError('test exception')
        with pytest.raises(RuntimeError):
            with a1:
                await a1.wait_events()
                a1.ready()

    async def test_context_manager_no_ready(self, caplog) -> None:
        """Test using :class:`resource.ResourceAllocation` as a context \
        manager when the user does not call :meth:`.ResourceAllocation.ready`."""
        with caplog.at_level(logging.WARNING, logger='katsdpsigproc.resource'):
            r = resource.Resource(None)
            a0 = r.acquire()
            with a0:
                pass
        assert caplog.record_tuples == [(
            'katsdpsigproc.resource',
            logging.WARNING,
            'Resource allocation was not explicitly made ready'
        )]


class TestJobQueue:
    @pytest.fixture(autouse=True)
    def setup(self, event_loop) -> None:
        self.jobs = resource.JobQueue()
        self.finished = [event_loop.create_future()
                         for i in range(5)]      # type: List[asyncio.Future[int]]
        self.unfinished = [event_loop.create_future()
                           for i in range(5)]    # type: List[asyncio.Future[int]]
        for i, future in enumerate(self.finished):
            future.set_result(i)

    def test_clean(self) -> None:
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[0])
        self.jobs.add(self.finished[1])
        self.jobs.clean()
        assert len(self.jobs._jobs) == 2

    async def _finish(self) -> None:
        for i, future in enumerate(self.unfinished):
            await asyncio.sleep(0.02)
            future.set_result(i)

    async def test_finish(self) -> None:
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[0])
        self.jobs.add(self.unfinished[1])
        self.jobs.add(self.unfinished[2])
        finisher = asyncio.ensure_future(self._finish())
        await self.jobs.finish(max_remaining=1)
        assert self.unfinished[0].done()
        assert self.unfinished[1].done()
        assert not self.unfinished[2].done()
        assert len(self.jobs) == 1
        await finisher

    def test_nonzero(self) -> None:
        assert not self.jobs
        self.jobs.add(self.finished[0])
        assert self.jobs

    def test_len(self) -> None:
        assert len(self.jobs) == 0
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[1])
        self.jobs.add(self.finished[1])
        assert len(self.jobs) == 3

    def test_contains(self) -> None:
        assert self.finished[0] not in self.jobs
        self.jobs.add(self.finished[0])
        assert self.finished[0] in self.jobs
        assert self.finished[1] not in self.jobs



import numpy as np
import katsdpsigproc.accel
from katsdpsigproc.accel import Operation, IOSlot, Dimension, build, roundup

pytest_plugins = ["katsdpsigproc.pytest_plugin"]

SOURCE = """
<%include file="/port.mako"/>

KERNEL void multiply(GLOBAL float *data, float scale) {
    data[get_global_id(0)] *= scale;
}
"""

class MultiplyTemplate:
    WGS = 32

    def __init__(self, context):
        self.program = build(context, '', source=SOURCE)

    def instantiate(self, queue, size, scale):
        return Multiply(self, queue, size, scale)


class Multiply(Operation):
    def __init__(self, template, queue, size, scale):
        super().__init__(queue)
        self.template = template
        self.kernel = template.program.get_kernel('multiply')
        self.scale = np.float32(scale)
        self.slots['data'] = IOSlot((Dimension(size, template.WGS),), np.float32)

    def _run(self):
        data = self.buffer('data')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [data.buffer, self.scale],
            global_size=(roundup(data.shape[0], self.template.WGS),),
            local_size=(self.template.WGS,)
        )


def test_multiply(context, command_queue):
    size = 53
    template = MultiplyTemplate(context)
    op = template.instantiate(command_queue, size, 4.0)
    op.ensure_all_bound()
    src = np.random.uniform(size=size).astype(np.float32)
    op.buffer('data').set(command_queue, src)
    op()
    dst = op.buffer('data').get(command_queue)
    np.testing.assert_array_equal(dst, src * 4.0)
